import os.path
import shutil
import sys
from typing import Optional

from johnsnowlabs import settings
from johnsnowlabs.auto_install.databricks.install_utils import (
    create_cluster,
    get_db_client_for_token,
    install_jsl_suite_to_cluster,
)
from johnsnowlabs.auto_install.install_software import check_and_install_dependencies
from johnsnowlabs.auto_install.jsl_home import (
    setup_jsl_home,
    get_install_suite_from_jsl_home,
)
from johnsnowlabs.auto_install.offline_install import get_printable_dependency_urls
from johnsnowlabs.auto_install.softwares import Software
from johnsnowlabs.py_models.jsl_secrets import JslSecrets
from johnsnowlabs.utils.enums import ProductName, PyInstallTypes, JvmHardwareTarget


def install(
    # -- JSL-Auth Flows --
    # Browser Auth
    browser_login: bool = True,
    force_browser: bool = False,
    # JWT Token Auth
    access_token: Optional[str] = None,
    # JSON file Auth
    json_license_path: Optional[str] = None,
    # Manual License specification Auth
    med_license: Optional[str] = None,
    enterprise_nlp_secret: Optional[str] = None,
    ocr_secret: Optional[str] = None,
    ocr_license: Optional[str] = None,
    fin_license: Optional[str] = None,
    leg_license: Optional[str] = None,
    aws_access_key: Optional[str] = None,
    aws_key_id: Optional[str] = None,
    # -- Databricks auth flows & Install Target --
    databricks_cluster_id: Optional[str] = None,
    databricks_token: Optional[str] = None,
    databricks_host: Optional[str] = None,
    databricks_password: Optional[str] = None,
    databricks_email: Optional[str] = None,
    # -- Install Params --
    # Install Target
    python_exec_path: str = sys.executable,
    venv_creation_path: Optional[str] = None,
    offline_zip_dir: Optional[str] = None,
    # Download Params
    offline: bool = False,
    install_optional: bool = True,
    install_licensed: bool = True,
    slim_install: bool = False,  # Only Downloads jars
    product: Optional[str] = ProductName.jsl_full.value,
    include_dependencies: bool = True,
    nlp: bool = True,
    spark_nlp: bool = True,
    visual: bool = False,
    # License usage & Caching
    local_license_number: int = 0,
    remote_license_number: int = 0,
    store_in_jsl_home: bool = True,
    # Install File Types
    hardware_platform: str = JvmHardwareTarget.cpu.value,
    py_install_type: str = PyInstallTypes.wheel.value,
    only_refresh_credentials: bool = False,
    refresh_install: bool = False,
    # -- Databricks Cluster Creation Params --
    block_till_cluster_ready=True,
    num_workers=1,
    cluster_name=settings.db_cluster_name,
    node_type_id=settings.db_node_type_id,
    driver_node_type_id=settings.db_driver_node_type,
    spark_env_vars=None,
    autotermination_minutes=60,
    spark_version=settings.db_spark_version,
    spark_conf=None,
    auto_scale=None,
    aws_attributes=None,
    ssh_public_keys=None,
    custom_tags=None,
    cluster_log_conf=None,
    enable_elastic_disk=None,
    cluster_source=None,
    instance_pool_id=None,
    headers=None,
):
    if refresh_install and os.path.exists(settings.root_dir):
        shutil.rmtree(settings.root_dir)
    # Input Validation
    py_install_type = PyInstallTypes.from_str(py_install_type)
    hardware_platform = JvmHardwareTarget.from_str(hardware_platform)
    product = Software.for_name(product)

    # Get Credentials from Auth Flow
    secrets: JslSecrets = JslSecrets.build_or_try_find_secrets(
        browser_login=browser_login,
        force_browser=force_browser,
        access_token=access_token,
        local_license_number=local_license_number,
        remote_license_number=remote_license_number,
        secrets_file=json_license_path,
        hc_license=med_license,
        hc_secret=enterprise_nlp_secret,
        ocr_secret=ocr_secret,
        ocr_license=ocr_license,
        aws_access_key=aws_access_key,
        aws_key_id=aws_key_id,
        return_empty_secrets_if_none_found=True,
        fin_license=fin_license,
        leg_license=leg_license,
        store_in_jsl_home=store_in_jsl_home,
    )
    if only_refresh_credentials:
        return

    if offline:
        # Offline Install
        get_printable_dependency_urls(
            secrets=secrets,
            jvm_install_type=hardware_platform,
            py_install_type=py_install_type,
        )
        if not offline_zip_dir:
            return

    if store_in_jsl_home and not offline:
        # Cache credentials, Wheels and Jars in ~/.johnsnowlabs
        setup_jsl_home(
            secrets=secrets,
            jvm_install_type=hardware_platform,
            py_install_type=py_install_type,
            refresh_install=refresh_install,
            visual=visual,
            nlp=nlp,
            spark_nlp=spark_nlp,
        )

    # Databricks Install
    if databricks_host and databricks_token and not offline:
        suite = get_install_suite_from_jsl_home(
            jvm_hardware_target=hardware_platform,
            visual=visual,
            nlp=nlp,
            spark_nlp=spark_nlp,
        )
        if databricks_cluster_id:
            install_jsl_suite_to_cluster(
                db=get_db_client_for_token(databricks_host, databricks_token),
                install_suite=suite,
                cluster_id=databricks_cluster_id,
            )

        else:
            return create_cluster(
                db=get_db_client_for_token(databricks_host, databricks_token),
                install_suite=suite,
                block_till_cluster_ready=block_till_cluster_ready,
                num_workers=num_workers,
                cluster_name=cluster_name,
                node_type_id=node_type_id,
                driver_node_type_id=driver_node_type_id,
                spark_env_vars=spark_env_vars,
                autotermination_minutes=autotermination_minutes,
                spark_version=spark_version,
                spark_conf=spark_conf,
                auto_scale=auto_scale,
                aws_attributes=aws_attributes,
                ssh_public_keys=ssh_public_keys,
                custom_tags=custom_tags,
                cluster_log_conf=cluster_log_conf,
                enable_elastic_disk=enable_elastic_disk,
                cluster_source=cluster_source,
                instance_pool_id=instance_pool_id,
                headers=headers,
            )

    # Local Py-Install
    elif not slim_install:
        check_and_install_dependencies(
            product=product,
            secrets=secrets,
            install_optional=install_optional,
            install_licensed=install_licensed,
            python_exec_path=python_exec_path,
            py_setup_dir=venv_creation_path,
            offline_zip_dir=offline_zip_dir,
            include_dependencies=include_dependencies,
            visual=visual,
            spark_nlp=spark_nlp,
            enterpise_nlp=nlp,
        )



from abc import ABCMeta

from johnsnowlabs.abstract_base.lib_resolver import (
    Py4JJslLibDependencyResolverABC,
    PyInstallTypes,
)
from johnsnowlabs.py_models.url_dependency import UrlDependency
from johnsnowlabs.utils.enums import (
    ProductName,
    SparkVersion,
    JvmHardwareTarget,
    LatestCompatibleProductVersion,
)


class HcLibResolver(Py4JJslLibDependencyResolverABC, metaclass=ABCMeta):
    has_py_install = True
    has_cpu_jars = True
    has_secret = True
    compatible_spark_versions = [SparkVersion.spark3xx.value]
    lib_version = LatestCompatibleProductVersion.healthcare.value
    product_name = ProductName.hc

    compatible_spark_to_jar_map = {
        SparkVersion.spark3xx: {
            JvmHardwareTarget.cpu: UrlDependency(
                url="https://pypi.johnsnowlabs.com/{secret}/spark-nlp-jsl-{lib_version}.jar",
                dependency_type=JvmHardwareTarget.cpu,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            )
        }
    }

    compatible_spark_to_py_map = {
        SparkVersion.spark3xx: {
            PyInstallTypes.wheel: UrlDependency(
                url="https://pypi.johnsnowlabs.com/{secret}/spark-nlp-jsl/spark_nlp_jsl-{lib_version}-py3-none-any.whl",
                dependency_type=PyInstallTypes.wheel,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
            PyInstallTypes.tar: UrlDependency(
                url="https://pypi.johnsnowlabs.com/{secret}/spark-nlp-jsl/spark-nlp-jsl-{lib_version}.tar.gz",
                dependency_type=PyInstallTypes.tar,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
        }
    }



import os
import shutil
import urllib
from pathlib import Path
from pprint import pprint
from typing import Union, List

import nbformat
import pandas as pd
from nbclient.exceptions import CellExecutionError
from nbconvert import PythonExporter

import johnsnowlabs.utils.testing.test_settings
from johnsnowlabs.utils.file_utils import path_tail
from johnsnowlabs.utils.py_process import str_to_file, log_multi_run_status
from johnsnowlabs.utils.testing.jsl_pre_processor import JslPreprocessor

Path(johnsnowlabs.utils.testing.test_settings.tmp_notebook_dir).mkdir(
    exist_ok=True, parents=True
)


def nbformat_to_ipynb(out_path, nb):
    with open(out_path, "w", encoding="utf-8") as f:
        nbformat.write(nb, f)


def ipynb_to_nbformat(in_path):
    with open(in_path, encoding="utf8") as f:
        return nbformat.read(f, as_version=4)


def get_all_nb_in_local_folder(p):
    ## filter all files ending in .ipynb
    return [f"{p}/{f}" for f in os.listdir(p) if ".ipynb" in f]


def convert_notebook(notebookPath):
    out_path = f"{johnsnowlabs.utils.testing.test_settings.tmp_notebook_dir}/{path_tail(notebookPath)}.nb_converted.py"
    with open(notebookPath, encoding="utf8") as fh:
        nb = nbformat.reads(fh.read(), nbformat.NO_CONVERT)
    exporter = PythonExporter()
    source, meta = exporter.from_notebook_node(nb)
    str_to_file(source, out_path)
    return out_path


def convert_all_notebooks(nb_folder):
    # Convert a folder which contains .ipynb into .py
    store_folder = f"{nb_folder}/nb_converted/"
    Path(store_folder).mkdir(parents=True, exist_ok=True)
    for nb_path in get_all_nb_in_local_folder(nb_folder):
        save_path = store_folder + nb_path.split("/")[-1] + ".py"
        convert_notebook(nb_path, save_path)


def test_ipynb(
    file_path_or_url: Union[List[str], str],
    use_i_py=True,
    model_cache_dir=None,
    out_dir=None,
    branch="master",
):
    """
    ref can be URL, PATH, DIR,
    or ref= `WORKSHOP` or `WORKSHOP-OS`  or `WORKSHOP-MED` , `WORKSHOP-LEG`, `WORKSHOP-FIN`
        for testing a specific sub-folder of the workshop
    use branch parameter to test a specific branch
    """
    if out_dir:
        os.chdir(out_dir)
    if isinstance(file_path_or_url, List):
        return test_list_of_ipynb(file_path_or_url, out_dir=out_dir)

    if not os.path.exists(file_path_or_url):
        # Remote handling
        if "http" and "//" in file_path_or_url:
            # URL
            file_name = file_path_or_url.split("/")[-1]
            print(f"Downloading {file_path_or_url} to  {file_name}")
            # Download the file from `url` and save it locally under `file_name`:
            with urllib.request.urlopen(file_path_or_url) as response, open(
                file_name, "wb"
            ) as out_file:
                shutil.copyfileobj(response, out_file)
                file_path_or_url = file_name

        elif "WORKSHOP" in file_path_or_url:
            # Workshop handing
            if not os.path.exists(
                johnsnowlabs.utils.testing.test_settings.workshop_cert_nb_folder
            ):
                # Clone repo
                cur_dir = os.getcwd()
                os.chdir(f"{johnsnowlabs.utils.testing.test_settings.tmp_notebook_dir}")
                os.system(
                    f"git clone {johnsnowlabs.utils.testing.test_settings.workshop_git} -b {branch}"
                )
                os.chdir(cur_dir)
            if "WORKSHOP-FIN" == file_path_or_url:
                return test_ipynb_folder(
                    johnsnowlabs.utils.testing.test_settings.workshop_fin_folder,
                    use_i_py=use_i_py,
                    model_cache_dir=model_cache_dir,
                    out_dir=out_dir,
                )
            if "WORKSHOP-LEG" == file_path_or_url:
                return test_ipynb_folder(
                    johnsnowlabs.utils.testing.test_settings.workshop_leg_folder,
                    use_i_py=use_i_py,
                    model_cache_dir=model_cache_dir,
                    out_dir=out_dir,
                )
            if "WORKSHOP-MED" == file_path_or_url:
                return test_ipynb_folder(
                    johnsnowlabs.utils.testing.test_settings.workshop_med_folder,
                    use_i_py=use_i_py,
                    model_cache_dir=model_cache_dir,
                    out_dir=out_dir,
                )
            if "WORKSHOP-PUB" == file_path_or_url:
                return test_ipynb_folder(
                    johnsnowlabs.utils.testing.test_settings.workshop_pub_folder,
                    use_i_py=use_i_py,
                    model_cache_dir=model_cache_dir,
                    out_dir=out_dir,
                )

            return pd.concat(
                [
                    test_ipynb_folder(
                        johnsnowlabs.utils.testing.test_settings.workshop_leg_folder,
                        use_i_py=use_i_py,
                        model_cache_dir=model_cache_dir,
                    ),
                    test_ipynb_folder(
                        johnsnowlabs.utils.testing.test_settings.workshop_fin_folder,
                        use_i_py=use_i_py,
                        model_cache_dir=model_cache_dir,
                    ),
                    test_ipynb_folder(
                        johnsnowlabs.utils.testing.test_settings.workshop_med_folder,
                        use_i_py=use_i_py,
                        model_cache_dir=model_cache_dir,
                    ),
                    test_ipynb_folder(
                        johnsnowlabs.utils.testing.test_settings.workshop_pub_folder,
                        use_i_py=use_i_py,
                        model_cache_dir=model_cache_dir,
                    ),
                ]
            )

    if os.path.isdir(file_path_or_url):
        # Folder
        return test_ipynb_folder(
            file_path_or_url,
            use_i_py=use_i_py,
            model_cache_dir=model_cache_dir,
            out_dir=out_dir,
        )

    if not os.path.isfile(file_path_or_url):
        raise ValueError(
            f"""Invalid target, must either be: 
                           1. Path to local Notebook
                           2. Path to local Notebook folder
                           3. URL to Remote Notebook (Make sure to use RAW github URL) 
                           4. WORKSHOP, WORKSHOP-OS, WORKSHOP-MED, WORKSHOP-LEG, WORKSHOP-FIN 
                           """
        )

    if not out_dir:
        out_dir = file_path_or_url + "_EXECUTED.ipynb"
    else:
        out_dir = f"{out_dir}/{path_tail(file_path_or_url)}"
    result = execute_nb_as_nb(file_path_or_url, out_dir)
    return result


def test_ipynb_folder(
    nb_folder,
    work_dir=os.getcwd(),
    log=True,
    model_cache_dir=None,
    use_i_py=True,
    out_dir=None,
):
    return test_list_of_ipynb(
        get_all_nb_in_local_folder(nb_folder),
        work_dir,
        log,
        model_cache_dir=model_cache_dir,
        out_dir=out_dir,
        use_i_py=use_i_py,
    )


def test_list_of_ipynb(
    nb_paths_or_urls,
    work_dir=os.getcwd(),
    log=True,
    model_cache_dir=None,
    use_i_py=True,
    out_dir=None,
):
    df = []
    for i, nb_path in enumerate(nb_paths_or_urls):
        print(f"Testing {i}/{len(nb_paths_or_urls)} {nb_path}")
        df.append(
            test_ipynb(
                file_path_or_url=nb_path,
                model_cache_dir=model_cache_dir,
                use_i_py=use_i_py,
                out_dir=out_dir,
            )
        )
    df = pd.DataFrame(df)
    if log:
        log_multi_run_status(df)
    return df


def execute_nb_as_nb(nb_path, out_file, work_dir=None):
    # Execute Notebook with custom JSL Preprocessor
    if not work_dir:
        work_dir = johnsnowlabs.utils.testing.test_settings.testing_work_dir
    Path(
        johnsnowlabs.utils.testing.test_settings.testing_work_dir,
    ).mkdir(exist_ok=True, parents=True)

    nb_name = nb_path.split("/")[-1]
    # Load file, Pre-Process and Write new File to Disk
    nb = ipynb_to_nbformat(nb_path)
    ep = JslPreprocessor(timeout=600, kernel_name="python3", allow_errors=True)
    failed = False
    exception_message = None
    try:
        ep.preprocess(nb, {"metadata": {"path": work_dir}})
        nbformat_to_ipynb(out_file, nb)
    except CellExecutionError as err:
        exception_message = err
        print(
            f"🚨 Failure Executing notebook cell \n{err}",
        )
        nbformat_to_ipynb(out_file, nb)
    except Exception as err:
        exception_message = err
        print(
            f"🚨 Unknown Exception while trying to run Notebook  \n{err}",
        )
        nbformat_to_ipynb(out_file, nb)
    if exception_message:
        failed = True

    bad_cells = []
    code_cells = list(filter(lambda x: x["cell_type"] == "code", nb["cells"]))
    for cell in code_cells:
        if "outputs" in cell:
            for cell_output in cell["outputs"]:
                if cell_output["output_type"] == "error":
                    bad_cells.append(
                        make_nbconvert_exec_log(
                            bad_cell_content=cell["source"],
                            error_output_type=cell_output["output_type"],
                            error_name=cell_output["ename"],
                            error_value=cell_output["evalue"],
                            error_traceback="\n".join(cell_output["traceback"]),
                        )
                    )

                    print(
                        f"\n🚨 Failure Executing Cell!  in {out_file}\n",
                    )
                    pprint(bad_cells[-1])
    success = len(bad_cells) == 0
    if not success:
        print(f"🚨 Failures in Notebook {nb_name} ")
    else:
        print(f"👌 No Failures in Notebook {nb_name} ")

    return make_nb_run_log(
        nb_name=nb_name,
        success=success,
        out_file=out_file,
        failed_cells_logs=bad_cells,
    )


def make_nbconvert_exec_log(
    bad_cell_content, error_output_type, error_name, error_value, error_traceback
):
    # Log for failed Cells
    return {
        "bad_cell_content": bad_cell_content,
        "error_output_type": error_output_type,
        "error_name": error_name,
        "error_value": error_value,
        "error_traceback": error_traceback,
    }
    pass


def make_nb_run_log(nb_name, success, out_file, failed_cells_logs):
    return {
        "success": success,
        "nb_name": nb_name,
        "output_notebook_path": out_file,
        "failed_cells_logs": failed_cells_logs,
    }



from .auto_install.health_checks.report import (
    check_health,
    list_remote_licenses,
    list_local_licenses,
)
from .utils.sparksession_utils import start
from .auto_install.install_flow import install
from johnsnowlabs.auto_install.databricks.work_utils import run_in_databricks
from johnsnowlabs import settings, viz, lab

from johnsnowlabs.abstract_base.lib_resolver import try_import_lib


if try_import_lib("sparknlp"):
    from sparknlp.base import *
    from sparknlp.annotator import *
    import sparknlp
    from sparknlp.pretrained import ResourceDownloader
    from sparknlp.training import *
    from sparknlp.functions import *
    from sparknlp.pretrained import PretrainedPipeline
    from sparknlp import annotation


if try_import_lib("pyspark"):
    from pyspark.sql import DataFrame
    import pyspark.sql.functions as F
    import pyspark.sql.types as T
    import pyspark.sql as SQL

    from pyspark import ml as ML
    import pyspark.ml.param.shared as _shared_pyspark_ml_param

    ML.param.shared = _shared_pyspark_ml_param

    from pyspark.sql import SparkSession
    from pyspark.ml import Pipeline, PipelineModel

if try_import_lib("warnings"):
    import warnings

    warnings.filterwarnings("ignore")

if try_import_lib("nlu"):
    from nlu import load, to_nlu_pipe, autocomplete_pipeline, to_pretty_df
    import nlu as nlu



# from johnsnowlabs import *
# spark = jsl.start()
from pyspark.ml import Pipeline
from sparknlp import DocumentAssembler
from sparknlp.annotator import ContextSpellCheckerModel, Tokenizer


# from johnsnowlabs import *
def run_test():
    # spark = jsl.start()
    documentAssembler = DocumentAssembler().setInputCol("text").setOutputCol("document")

    tokenizer = Tokenizer().setInputCols(["document"]).setOutputCol("token")

    spellModel = (
        ContextSpellCheckerModel.pretrained(
            "spellcheck_clinical", "en", "clinical/models"
        )
        .setInputCols("token")
        .setOutputCol("checked")
    )

    pipeline = Pipeline(stages=[documentAssembler, tokenizer, spellModel])

    empty = spark.createDataFrame([[""]]).toDF("text")
    example = spark.createDataFrame(
        [
            [
                "Witth the hell of phisical terapy.",
            ]
        ]
    ).toDF("text")
    pipeline.fit(empty).transform(example).show()



from typing import Dict, Tuple

from colorama import Fore

from johnsnowlabs import settings
from johnsnowlabs.abstract_base.software_product import AbstractSoftwareProduct
from johnsnowlabs.auto_install.softwares import Software
from johnsnowlabs.py_models.jsl_secrets import LicenseInfos
from johnsnowlabs.utils.enums import ProductName
from johnsnowlabs.utils.my_jsl_api import get_access_key_from_browser, get_user_licenses


def check_health(check_install=True):
    # Print status of installations and licenses
    install_status: Dict[AbstractSoftwareProduct:bool] = {}
    health_check: Dict[AbstractSoftwareProduct:bool] = {}
    license_check: Dict[str, Tuple[bool, bool]] = {}
    print(f'{"_" * 10}Installation check results{"_" * 10}')
    for product in ProductName:
        if check_install:
            product = Software.for_name(product)
            if not product or not product.pypi_name:
                continue
            install_status[product] = (
                product.check_installed() and product.check_installed_correct_version()
            )
            if not product.check_installed():
                print(f"{product.logo + product.name} is not installed 🚨")
            elif not product.check_installed_correct_version():
                print(
                    f"{product.logo + product.pypi_name + Fore.LIGHTRED_EX}=={product.get_installed_version() + Fore.RESET} "
                    f"is installed but should be {product.pypi_name}=={Fore.LIGHTGREEN_EX + product.latest_version.as_str() + Fore.RESET} 🚨 To fix run:"
                )

                if product.licensed:
                    print(
                        f"{Fore.LIGHTGREEN_EX}nlp.install(){Fore.RESET} while authorizing to upgrade {product.logo + product.pypi_name}"
                    )
                else:
                    print(
                        f"{Fore.LIGHTGREEN_EX}python -m pip install {product.pypi_name}=={product.latest_version.as_str()} --upgrade{Fore.RESET}"
                    )
            else:
                print(
                    f"{product.logo + product.pypi_name}=={product.get_installed_version()} "
                    f"{Fore.LIGHTGREEN_EX}is correctly installed! ✅{Fore.RESET}"
                )

        if health_check:
            health_check[product] = product.health_check()


def list_remote_licenses():
    access_token = get_access_key_from_browser()
    licenses = get_user_licenses(access_token)

    print("Your Remote licenses in https://my.johnsnowlabs.com:")
    for i, lic in enumerate(licenses):
        softwares = [Software.for_name(p.value) for p in lic.products]
        print(
            f'Remote License Number {i} Has access to: {", ".join([s.logo + s.name for s in softwares])}.\n'
            f"Extra info for License {i}: EndDate={lic.endDate}, Type={lic.type}, Platform={lic.platform}, Id={lic.id} "
        )

        print("_" * 50)


def list_local_licenses():
    print(f"Your licenses in {settings.license_dir}")
    licenses = LicenseInfos.from_home()
    if not licenses:
        raise Exception("You have no local licenses")
    i = 0
    for file, lic in licenses.infos.items():
        softwares = [Software.for_name(p.value) for p in lic.products]
        print(
            f'Your local license in {i} Has access to: {", ".join([s.logo + s.name for s in softwares])}.\n'
            f"Extra info for License {i}: File={file} Id={lic.id}"
        )
        i += 1
        print("_" * 50)
